import { toChainId } from "@wormhole-foundation/sdk-base";
import { SolanaPlatform, SolanaAddress, } from "@wormhole-foundation/sdk-solana";
import { AddressLookupTableProgram, Keypair, PublicKey, SystemProgram, TransactionMessage, VersionedTransaction, } from "@solana/web3.js";
import { createAssociatedTokenAccountIdempotentInstruction, createTransferInstruction, getAssociatedTokenAddressSync, } from "@solana/spl-token";
import { BN, Program } from "@coral-xyz/anchor";
import { ExampleNttWithExecutorIdl, } from "../idl/executor/example_ntt_with_executor.js";
import { ExampleNttSvmLutIdl, } from "../idl/executor/example_ntt_svm_lut.js";
import { chainToBytes } from "../lib/utils.js";
export class SolanaNttWithExecutor {
    network;
    chain;
    connection;
    contracts;
    nttWithExecutorProgramId;
    nttLutProgramId;
    executorProgramId;
    constructor(network, chain, connection, contracts) {
        this.network = network;
        this.chain = chain;
        this.connection = connection;
        this.contracts = contracts;
        this.nttWithExecutorProgramId = new PublicKey("execXUrAsMnqMmTHj5m7N1YQgsDz3cwGLYCYyuDRciV");
        this.nttLutProgramId = new PublicKey("1uteB5DZdNfns9B12rgGf5msKh1d7FbbkvciWmhsZiC");
        this.executorProgramId = new PublicKey("nex1gkSWtRBheEJuQZMqHhbMG5A45qPU76KqnCZNVHR");
    }
    static async fromRpc(connection, config) {
        const [network, chain] = await SolanaPlatform.chainFromRpc(connection);
        const conf = config[chain];
        if (conf.network !== network)
            throw new Error(`Network mismatch: ${conf.network} != ${network}`);
        return new SolanaNttWithExecutor(network, chain, connection, conf.contracts);
    }
    async *transfer(sender, destination, amount, quote, ntt, wrapNative = false) {
        const senderPk = new SolanaAddress(sender).unwrap();
        const options = { queue: false, automatic: false, wrapNative };
        const outboxItem = Keypair.generate();
        const txs = ntt.transfer(sender, quote.remainingAmount, destination, options, outboxItem);
        for await (const tx of txs) {
            if (tx.description === "Ntt.Transfer") {
                const luts = [];
                try {
                    luts.push(await ntt.getAddressLookupTable());
                }
                catch (e) {
                    console.debug(e);
                }
                const message = TransactionMessage.decompile(tx.transaction.transaction.message, { addressLookupTableAccounts: luts });
                if (quote.referrerFee > 0n) {
                    const referrer = new PublicKey(quote.referrer.address.toString());
                    const { mint } = await ntt.getConfig();
                    const referrerAta = getAssociatedTokenAddressSync(mint, referrer, true);
                    const senderAta = getAssociatedTokenAddressSync(mint, senderPk, true);
                    const referrerAtaAccount = await this.connection.getAccountInfo(referrerAta);
                    if (!referrerAtaAccount) {
                        message.instructions.push(createAssociatedTokenAccountIdempotentInstruction(senderPk, referrerAta, referrer, mint));
                    }
                    message.instructions.push(createTransferInstruction(senderAta, referrerAta, senderPk, quote.referrerFee));
                }
                const nttWithExecutorProgram = new Program(ExampleNttWithExecutorIdl, this.nttWithExecutorProgramId, { connection: this.connection });
                const nttProgramId = new PublicKey(this.contracts.ntt.manager);
                const nttPeer = PublicKey.findProgramAddressSync([Buffer.from("peer"), chainToBytes(destination.chain)], nttProgramId)[0];
                message.instructions.push(await nttWithExecutorProgram.methods
                    .relayNttMesage({
                    execAmount: new BN(quote.estimatedCost.toString()),
                    recipientChain: toChainId(destination.chain),
                    signedQuoteBytes: Buffer.from(quote.signedQuote),
                    relayInstructions: Buffer.from(quote.relayInstructions),
                })
                    .accounts({
                    payer: senderPk,
                    payee: new PublicKey(quote.payeeAddress),
                    nttProgramId,
                    nttPeer,
                    nttMessage: outboxItem.publicKey,
                    executorProgram: this.executorProgramId,
                })
                    .instruction());
                if (luts.length === 0) {
                    console.debug("no manager lookup table found, checking helper program");
                    const nttSvmLutProgram = new Program(ExampleNttSvmLutIdl, this.nttLutProgramId, { connection: this.connection });
                    const lutPointerAddress = PublicKey.findProgramAddressSync([Buffer.from("lut"), nttProgramId.toBuffer()], nttSvmLutProgram.programId)[0];
                    let lutPointer = 
                    // @ts-ignore
                    await nttSvmLutProgram.account.lut.fetchNullable(lutPointerAddress);
                    if (!lutPointer) {
                        console.debug("no helper program lookup table found, initializing...");
                        const [nttConfigPDA] = await PublicKey.findProgramAddressSync([Buffer.from("config")], nttProgramId);
                        const [authorityPDA] = await PublicKey.findProgramAddressSync([Buffer.from("lut_authority")], nttSvmLutProgram.programId);
                        const recentSlot = (await this.connection.getSlot("finalized")) - 10;
                        const [lutAddressPDA] = await PublicKey.findProgramAddressSync([
                            authorityPDA.toBuffer(),
                            new BN(recentSlot).toArrayLike(Buffer, "le", 8),
                        ], AddressLookupTableProgram.programId);
                        const [lutPDA] = await PublicKey.findProgramAddressSync([Buffer.from("lut"), nttProgramId.toBuffer()], nttSvmLutProgram.programId);
                        const ix = await nttSvmLutProgram.methods
                            .initializeLut(new BN(recentSlot))
                            .accounts({
                            nttProgramId,
                            nttConfig: nttConfigPDA,
                            authority: authorityPDA,
                            lutAddress: lutAddressPDA,
                            lut: lutPDA,
                            lutProgram: AddressLookupTableProgram.programId,
                            systemProgram: SystemProgram.programId,
                            payer: senderPk,
                        })
                            .instruction();
                        const lutInitMessage = new TransactionMessage({
                            payerKey: senderPk,
                            recentBlockhash: (await this.connection.getLatestBlockhash("finalized")).blockhash,
                            instructions: [ix],
                        }).compileToV0Message();
                        yield ntt.createUnsignedTx({ transaction: new VersionedTransaction(lutInitMessage) }, "NttSvmLut.InitializeLut");
                        console.debug(`initialized lookup table: ${tx}`);
                        let retries = 0;
                        while (!lutPointer && retries < 10) {
                            // wait for lut to warm up
                            await new Promise((resolve) => setTimeout(resolve, 2000));
                            lutPointer =
                                // @ts-ignore
                                await nttSvmLutProgram.account.lut.fetchNullable(lutPointerAddress);
                            retries++;
                        }
                    }
                    const response = await this.connection.getAddressLookupTable(lutPointer.address);
                    if (!response.value) {
                        throw new Error("unable to fetch lookup table");
                    }
                    luts.push(response.value);
                }
                tx.transaction.transaction.message = message.compileToV0Message(luts);
                yield tx;
            }
            else {
                yield tx;
            }
        }
    }
    static associatedTokenAccountMinRent = undefined;
    async estimateMsgValueAndGasLimit(recipient) {
        let msgValue = 0n;
        // there could be up to 2 transactions to post the signatures/vaa
        // TODO: check that this is right cost in mainnet for signatures accounts
        msgValue += 2n * (15001n + 1200000n); // signatures accounts, e.g. https://explorer.solana.com/tx/2bTVAkZS1MmVHejY5bkNYsgqHN9RYuJYEXkH6m62xEgP9oeqAHCjTng67YWQGpgibQXUKTEFqYZCwwHFhWmtgtbf?cluster=devnet
        msgValue += 5001n + 3100000n; // core bridge post vaa account, e.g. https://explorer.solana.com/tx/5Rhw75qBdt35L7Npuj1Uc7Ua824jyXPVJUG4dw33hN9getuzKGniGYwcgJpJ69QsWAmQu4mChJMfDnFyPzibCzUp?cluster=devnet
        msgValue += 5001n + 5000000n; // receive wormhole message accounts, e.g. https://explorer.solana.com/tx/3R1cj71FY2uWYygnt5xQG2ToPviVnJhNrCarguAcEFk3GSCVRTFgtwvNHabtg324NKnMjQidgsg4wg954EnuZgsN?cluster=devnet
        msgValue += 5001n; // release has no accounts, unless sending to a new ATA, e.g. https://explorer.solana.com/tx/5YFf7DxZfuJvhm38qoZPGsWXmGuCUUYSiYozDgUZUa5joEusiDwBCyBXUfH1sqTaxq9Gf9c4P5FfJSkLcFXbWpJe?cluster=devnet
        if (recipient) {
            const recipientPk = new PublicKey(recipient.address.toString());
            const ata = getAssociatedTokenAddressSync(new SolanaAddress(this.contracts.ntt.token).unwrap(), recipientPk, true);
            if ((await this.connection.getAccountInfo(ata)) === null) {
                if (!SolanaNttWithExecutor.associatedTokenAccountMinRent) {
                    SolanaNttWithExecutor.associatedTokenAccountMinRent = BigInt(await this.connection.getMinimumBalanceForRentExemption(165) // ATA is 165 bytes
                    );
                }
                msgValue += SolanaNttWithExecutor.associatedTokenAccountMinRent;
            }
        }
        return { msgValue, gasLimit: 250000n };
    }
}
//# sourceMappingURL=nttWithExecutor.js.map